from typing import MutableMapping, Mapping, Dict, Callable, Optional, Any
from .storage import Storage


class MemoryStorage(Storage):
    def __init__(self):
        """
        A memory storage object for very fast reading and writing
        """

        self._memory = None

    def read(self) -> Optional[MutableMapping[Any, Any]]:
        return self._memory

    def write(self, data: Mapping[Any, Any]) -> None:
        self._memory = data


class Piuma:
    def __init__(self, storage: Storage = MemoryStorage()):
        """
        Piuma `__init__` method, initializes the database by setting the
        storage object and the `_next_id`.

        Parameteres
        -----------
        storage: Storage, optional
            The storage object where the database is stored.
        """

        self._storage = storage
        self._next_id = int()

    def insert(self, value: Mapping, id: int = int()) -> int:
        """
        Inserts a new document into the database.

        Parameters
        ----------
        value: Mapping
            The value of the document that will be inserted. The value can take
            in any `Mapping` type, the simplest example is `dict`.
        id: int, optional
            The id of the document. This parameter is optional, if no id is
            passed in an id will be generated.

        Returns
        -------
        int
            returns the id of the inserted document
        """

        if not id:
            id = self._get_next_id()
        else:
            self._next_id = id

        def updater(database: MutableMapping[Any, Any]):
            """
            A common update method that specifies how to modify the database.
            Inserts a new document into the database.
            """

            if id not in database:
                database[id] = value
            else:
                raise KeyError("Document with the specified id already exists")

        self._update_database(updater)

        return id

    def get(self, id: int) -> Optional[Dict]:
        """
        Returns a document with a specified id.

        Parameters
        ----------
        id: int
            The id of the document that will be searched for

        Returns
        -------
        dict
            Returns the document that has specified id if it exists, if it does
            not exist then it returns none.
        """

        database = self._storage.read()

        if not database:
            database = {}

        return database[id] if id in database else None

    def remove(self, id: int) -> None:
        """
        Removes the document with the specified id

        Parameters
        ----------
        id: int
            The id of the document that will be searched for
        """

        def updater(database: MutableMapping[Any, Any]):
            """
            A common update method that specifies how to modify the database.
            Removes the document with the specified id
            """

            database.pop(id)

        self._update_database(updater)

    def update(self, value: Mapping, id: int) -> None:
        """
        Updates the document with the specified id

        Parameters
        ----------
        value: Mapping
            The value that will replace the previous value of the specified
            document. The value can take in any `Mapping` type, the simplest
            example is `dict`.
        id: int
            The id of the document that will be searched for
        """

        def updater(database: MutableMapping[Any, Any]):
            """
            A common update method that specifies how to modify the database.
            If the document with the specified id exists, update it.
            """

            if id in database:
                database[id] = value
            else:
                # TODO should this not return an error or `None`?
                raise KeyError(id)

        self._update_database(updater)

    def all(self) -> Optional[MutableMapping[Any, Any]]:
        return self._storage.read()

    def _get_next_id(self) -> int:
        """
        Returns the next available id in the database
        """

        if self._next_id is not None:
            self._next_id += 1

            return self._next_id

        database = self._storage.read()

        # if the database is empty then _next_id is 1
        if database:
            # the _next_id is 1 more than the biggest id
            self._next_id = max(key for key in database.keys()) + 1
        else:
            self._next_id = 1

        return self._next_id

    def _update_database(self, updater: Callable) -> None:
        """
        Run `updater` method to update the database and initialize it if it is
        not already

        Parameters
        ----------
        updater: Callable
            A method that specifies how to update the database
        """

        database = self._storage.read()

        # Initializes database if empty
        if not database:
            database = {}

        # Applies changes to the database
        updater(database)

        self._storage.write(database)



import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Mapping, MutableMapping, Optional


class Storage(ABC):
    """
    An abstract class for all storages. This class ensures that any storage
    class has a read and a write method.
    """

    @abstractmethod
    def read(self) -> Optional[MutableMapping[Any, Any]]:
        """
        The read method should return the entire database, can optionally
        return None for initialization. If any data needs to be deserialized,
        it can be done here.
        """

        raise NotImplementedError("Not Overwritten")

    @abstractmethod
    def write(self, data: Mapping[Any, Any]) -> None:
        """
        The write method take in the entire database and write it. If any data
        needs to be serialized, it can be done here.
        """

        raise NotImplementedError("Not Overwritten")

    def close(self):
        """
        Optional Close mtheod to close any lingering handles
        """

        pass


class Cache:
    def __init__(self, storage: Storage, max_modifications: int = 500):
        """
        A Cache object that will act as a middleware by storing the data in
        memory and then writing to the storage object after a number of
        modifications.
        """

        self._storage = storage

        self._memory = None

        # Number of modifications before it writes to the storage
        self._max_modifications = max_modifications
        # Number of modifications
        self._modifications = 0

    def read(self):
        # Returns cached memory
        return self._memory

    def write(self, data):
        # Sets memory cache to data
        self._memory = data

        self._modifications += 1

        # If the modifications surpass the max modifications write to the file
        if self._modifications > self._max_modifications:
            self.flush()

    def flush(self):
        # Writes to file
        self._storage.write(self._memory)

        # Resets modifications
        self._modifications = 0

    def close(self):
        self.flush()

        self._storage.close()


class JSONStorage(Storage):
    def __init__(self, path: str, makedirs: bool = False):
        """
        If the database file does not exist it is created. A handle object is
        created to talk to the database file

        Parameters
        ----------
        path: str
            The path of the database file ie, "data/database.json".
        makedirs: bool
            Make directories leading to the database file if they do not exist.
        """

        if not os.path.exists(path):
            if makedirs:
                os.makedirs(os.path.dirname(path))

            with open(path, "a"):
                pass

        self._handle = open(path, mode="r+", encoding="utf-8")

    def read(self) -> Optional[MutableMapping[int, Dict[Any, Any]]]:
        # Gets the size so no error is returned if an empty file is read
        self._handle.seek(0, os.SEEK_END)
        size = self._handle.tell()

        if not size:
            return None

        else:
            # Moves cursor to the beginning reads and deserializes the raw file
            self._handle.seek(0)
            raw = json.load(self._handle)

            # Converts all the ids from `str` to int
            database = {int(id): document for id, document in raw.items()}

            return database

    def write(self, data: Mapping[int, Dict[Any, Any]]) -> None:

        if not data:
            data = {}

        # Move cursor to the beginning
        self._handle.seek(0)

        # Serialize the file
        serialized = json.dumps(data)

        self._handle.write(serialized)

        # Makes sure the file is written to
        self._handle.flush()
        os.fsync(self._handle.fileno())

        # If the file has gotten shorter it removes any extra pieces
        self._handle.truncate()

    def close(self):
        self._handle.close()



from setuptools import setup, find_packages

LONG_DESCRIPTION = """
Piuma
=====

Piuma is an no-headache ultra-lightweight local document database written
purely in python. Piuma is designed without just the **absoulute bare**
essentials in mind. In the core version of Piuma there is no query language or
any advanced search/modify functions. In its current version Piuma only stores
data in memory, I plan to add local storage subpackage hopefully in the near
future. This database should not be taken too seriously and should really only
be used for small projects and testing.

Quick Links
-----------
- Github: https://github.com/emit07/piuma
- Documentation: https://piuma.readthedocs.io/

Why you *should* use Piuma?
~~~~~~~~~~~~~~~~~~~~~~~~~~~

- **Light and Small:** Piuma's core is written with one very small file and can
    be sub millisecond fast because of its memory storage
- **Easy to Use:** Piuma runs out of the box and is easy to use and learn.
    Piuma also requires no dependencies or external servers.
- **Experimenting:** If you are trying something new and you need a fast
    database that just works while testing then Piuma is a good choice.
- **No Dependencies:** Piuma is written with no external dependecies.

Why you *should not* use Piuma?
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- **Big Database:** Piuma was not designed to be used as a big database; Piuma
    only runs *only* in memory (for now) so the limit of how big it could be
    comes much quicker than other database
- **ACID Guarantees:** Because of Piumas small nature it does not have room for
    ACID_ Guarantees
- **Query Language:** Piuma does not have a query language (though I currently
    working on a query submodule)
- **Storage:** While Piuma supports interchangeable storage, it only has memory
    storage built in

Why is it called Piuma?
~~~~~~~~~~~~~~~~~~~~~~~

Piuma is the Italian word for feather and (from what I feel) the word feather
can convay lightness.

.. _ACID: https://en.wikipedia.org/wiki/ACID
"""

setup(
    name="piuma",
    version="1.1.2",
    license="gpl-3.0",
    author="Alessandro De Leo",
    author_email="emit07@protonmail.com",
    description="An ultra-lighweight document oriented database",
    long_description_content_type="text/x-rst",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    keywords=['python', 'database', 'document oriented'],
    url="https://github.com/emit07/piuma",
)
